/***********************************************
logistic_marginal.sas

Macro permettant de calculer facilement
les effets marginaux moyens dans un modle
de rgression logistique dichotomique : 
- la macro %marginal est directement extraite
de (Afsa, 2016);
- la macro %logistic_marginal est une surcouche
visant  faciliter l'utilisation de la macro
%marginal.

Rfrence : 
- Afsa C. (2016) Le modle Logit. Thorie et application, document de travail de l'Insee


Martin Chevalier
Formation Charg d'tudes statistiques
CEPE Ensae-Ensai Formation Continue
1-3 mars 2017
***********************************************/


/***********************************************
				MACRO %marginal
***********************************************/

%macro marginal(tab_ent=,x=,param_ent=,var_qual=,ponder=);
/* etape prealable, executee si il y a une variable de pondration */
%if &ponder ne %then %do;
proc summary data=&tab_ent(keep=&ponder);
var &ponder;
output out=poidsm(keep=poidsm) mean=poidsm;
run;
data &tab_ent(drop=poidsm);
if _n_=1 then set poidsm;
set &tab_ent;
poids=&ponder/poidsm;
run;
proc delete data=poidsm;run;
%end;
/* calcul et impression de l'effet marginal */
proc iml;
start lecture;
use &tab_ent;read all var{&x} into x;
use &param_ent;read all var{intercept &x}
where (_type_="PARMS") into b;
use &param_ent;read all var{intercept &x}
where (_type_="COV") into cov;
%if &ponder ne %then %do;
use &tab_ent;read all var{poids} into poids;
%end;
n=nrow(x);
x=j(n,1,1)||x;
beta=t(b);
explic={&x};
qual={&var_qual};
finish lecture;
start effet;
/* on repre le rang, dans la liste &x des variables du modele, de la
variable qualitative &var_qual (ou de la 1ere variable de la liste
&var_qual s'il s'agit d'une variable polytomique) */
r=0;
do q=1 to ncol(explic);
if explic[q]=qual[1] then r=q;
end;
/*** initialisation des grandeurs utilisees ... */
/* ... pour le calcul des effets marginaux */
delta=j(n,ncol(qual),0);
delta_moy=j(ncol(qual),1,0);
/* ... pour le calcul des ecarts-types */
gradi=j(1,ncol(x),0);
grad=j(ncol(qual),ncol(x),0);
sigma=j(ncol(qual),1,0);
p_value=j(ncol(qual),1,0);
/*** calcul des effets marginaux */
/* situation o var_qual=0 */
x[,r+1:r+ncol(qual)]=j(n,ncol(qual),0);
x_0=x;
g0=1/(1+exp(-x_0*beta));
%if &ponder ne %then %do;
g0=g0#poids;
%end;
/* situation o var_qual=1 */
do j=1 to ncol(qual);
x=x_0;
x[,r+j]=j(n,1,1);
g=1/(1+exp(-x*beta));
%if &ponder ne %then %do;
g=g#poids;
%end;
delta[,j]=g-g0;
delta_moy=t(delta[+,]*100/n);
/* calcul de l'cart-type */
do i=1 to n;
gradi=x[i,]#(g[i]#(1-g[i])-g0[i]#(1-g0[i]));
gradi[1,r+j]=g[i]#(1-g[i]);
grad[j,]=grad[j,]+gradi[1,];
end;
grad[j,]=grad[j,]/n;
sigma[j]=sqrt(grad[j,]*cov*t(grad[j,]))*100;
end;
p_value=2*(1-probnorm(abs(delta_moy)/sigma));
finish effet;
/*MC 
start impress;
delta_moyc=char(delta_moy,10,4);
sigmac=char(sigma,10,4);
p_valuec=char(p_value,12,4);
do j=1 to ncol(qual);
if p_value[j]<0.0001 then p_valuec[j]="
< 0.0001";
end;
effetc=delta_moyc||sigmac||p_valuec;
noms_ligne=rowcat(t({&var_qual})||j(ncol(qual),1,"
"));
noms_col={"effet marginal"," cart_type"," significativit"};
mattrib effetc rowname=noms_ligne
colname=noms_col
label=" ";
print "Effet marginal de '&var_qual'";
print effetc;
finish impress;*/
run lecture;
run effet;
effet=delta_moy||sigma||p_value;
create _lm_effet_ from effet;
append from effet;
close _lm_effet_;
quit;
/*%if &syserr ne 0 %then %do;
data _message_;
message="Attention ! Erreur !";
run;
proc print data=_message_ noobs;
var message;
run;
proc delete data=_message_;run;
%end;
Fin MC*/
%mend;



/***********************************************
			MACRO %logistic_marginal
***********************************************/

/*Les paramtres de la macro %logistic_marginal
sont analogues  ceux d'une rgression logistique 
classique, sinon que l'option PARAM = REF doit 
imprativement tre prcise dan l'instruction 
CLASS.*/

%MACRO logistic_marginal(
	DATA =, 
	CLASS =, 
	MODEL =, 
	WEIGHT =,
	OUTPUT =
);

%LET option_notes = %SYSFUNC(GETOPTION(NOTES));
OPTIONS NONOTES;

/*Etape 0 : Contrles minimaux*/
%IF %NRBQUOTE(&CLASS) = %STR() %THEN %DO; 
	%PUT %NRSTR(ERROR: L%'instrution CLASS doit tre renseigne.);
	%ABORT;
%END;
%IF %INDEX(%NRBQUOTE(%SYSFUNC(COMPRESS(&CLASS))),%STR(PARAM=REF)) = 0 %THEN %DO; 
	%PUT %NRSTR(ERROR: L%'option PARAM = REF de l%'instruction CLASS doit tre utilise.);
	%ABORT;
%END;
PROC LOGISTIC DATA = &DATA NOPRINT;
	CLASS &CLASS;
	MODEL &MODEL;
	%IF %NRBQUOTE(&WEIGHT) NE %STR() %THEN WEIGHT &WEIGHT;;
RUN;
%IF &syserr %THEN %GOTO erreur;


/*Etape 1 : Estimation du modle logistique*/
ODS OUTPUT ClassLevelInfo  = _lm_class_;
PROC LOGISTIC DATA = &DATA;
	CLASS &CLASS;
	MODEL &MODEL;
	%IF %NRBQUOTE(&WEIGHT) NE %STR() %THEN WEIGHT &WEIGHT;;
RUN;

/*Etape 2 : Reconstitution de la matrice des variables
indicatrices*/
DATA _lm_class_;
	SET _lm_class_; 
	LENGTH var $ 32;
	RETAIN var '';
	IF class NE '' THEN var = UPCASE(class);
	IF SUM(OF X:) = 1;
RUN;
DATA _lm_class_; 
	SET _lm_class_; 
	LENGTH indic $ 32  code $ 1000;
	indic = '_V'||COMPRESS(PUT(_N_, BEST.))||'_';
	code = 'IF '||TRIM(var)||' NE "" THEN '||TRIM(indic)||' = ( TRIM('||TRIM(var)||') = "'||TRIM(Value)||'");';
RUN;
PROC SQL NOPRINT;
	SELECT code, indic INTO :code SEPARATED BY ' ', :qual SEPARATED BY ' ' FROM _lm_class_;
QUIT;
DATA _lm_data_; 
	SET &DATA; 
	&code;
RUN;

/*Etape 3 : Rcriture de l'instruction MODEL*/
%LET MODEL2 = %QSCAN(&MODEL, 1, %STR(=));
%LET EXPL = %SYSFUNC(COMPBL(%QSCAN(&MODEL, 2, %STR(=))));
DATA _lm_expl_; 
	LENGTH expl $ 1000 var $ 32;
	expl = "%NRBQUOTE(&expl)";
	numVar = 1;
	DO WHILE(SCAN(expl, numVar, ' ') NE '');
		var = UPCASE(SCAN(expl, numVar, ' '));
		OUTPUT;
		numVar = numVar + 1;
	END;
RUN;
PROC SQL NOPRINT;
	CREATE TABLE _lm_expl2_ AS
		SELECT a.var, a.numVar, COALESCE(b.indic, a.var) AS newvar LENGTH 32
		FROM _lm_expl_ AS a
			LEFT JOIN _lm_class_ AS b ON a.var = b.var
		ORDER BY numVar, indic
	;
	SELECT newvar INTO :EXPL2 SEPARATED BY ' ' FROM _lm_expl2_;
QUIT;
%LET MODEL2 = &MODEL2 = &EXPL2;

/*Etape 4 : Estimation du modle logistique adapt pour 
la macro %marginal*/
PROC LOGISTIC DATA = _lm_data_ COVOUT OUTEST = _lm_param_ NOPRINT;
	MODEL &MODEL2;
	%IF %NRBQUOTE(&WEIGHT) NE %STR() %THEN WEIGHT &WEIGHT;;
RUN;

/*Etape 5 : Excution de la macro %marginal*/
%marginal(
	tab_ent = _lm_data_, x=&EXPL2, param_ent=_lm_param_, var_qual = &qual 
	%IF %NRBQUOTE(&WEIGHT) NE %STR() %THEN , ponder = &WEIGHT;
);

/*Etape 6 : Rcupration des rsultats et affichage*/
DATA _lm_effet_;
	SET _lm_effet_;
	LENGTH indic $ 32;
	indic = '_V'||COMPRESS(PUT(_N_, BEST.))||'_';
RUN;
PROC SQL;
	CREATE TABLE _lm_effet_ AS
		SELECT b.var LABEL "Variable", b.VALUE LABEL "Modalite"
			, a.COL1 FORMAT 8.4 LABEL "Effet marginal moyen"
			, a.COL2 FORMAT 8.4 LABEL "Ecart-type"
			, a.COL3 FORMAT PVALUE6.4 LABEL "P-valeur"
		FROM _lm_effet_ AS a LEFT JOIN _lm_class_ AS b ON a.indic = b.indic
	;
	SELECT * FROM _lm_effet_;
QUIT;

/*Etape 7 : Export si OUTPUT est renseign*/
%IF %NRBQUOTE(&OUTPUT) NE %STR() %THEN %DO;
	PROC SQL;
		OPTIONS NOTES;
		CREATE TABLE &OUTPUT AS 
			SELECT * FROM _lm_effet_
		;
		OPTIONS NONOTES;
	QUIT;
%END;
OPTIONS NOTES;
%PUT %NRSTR(NOTE: Macro %logistic_marginal excute avec succs.);
OPTIONS NONOTES;

%GOTO fin;

%erreur: 
%PUT %NRSTR(ERROR: Interruption anormale de la macro %logistic_marginal.);

%fin:

/*Etape 7 : Nettoyage et restauration des messages d'information*/
PROC DATASETS LIBRARY = WORK NOLIST; 
	DELETE _lm_:;
RUN;QUIT;
OPTIONS &option_notes;


%MEND;
